/*
**  routines to hande code editing by an external editor
**
**  Development History:
**      who                        when            why
**      muquit@muquit.com          May-16-2001     dont' like to edit code in the
**                                                 dialog.
**	alvinbeach@hotmail.com         Jan-30-2005     Updated to work with fltk 1.1.6 and fluid 1.3
**                                                 Only tested on Linux. I don't have Winodws
**                                                 so the Win32 code hasn not been tested.
**                                                 Also, the statusbar code has been commeted as to
**                                                 comply with the fluid.
**
**  alvinbeach@hotmail.com         Jan-31-2005     Updated to support fltk 2.0.
**                                                 This is not a clean patch. The code needs to be cleaned up
**                                                 but it does what it is suppose to do.
*/


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef WIN32
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#endif  /* ! WIN32 */

#include "function_panel.h"
#include "code_editor.h"
#include <FL/x.H>

#ifdef WIN32
#include <process.h>
#include <FL/fl_message.H>
#endif /* WIN32 */

// #define DEBUG 1

/**** globals in this file **** -- starts -- */

static char
    s_tmpfile[2048];    /* tmp file path */

static int
    s_editor_running=0; /* if editor is running */

/**** globals in this file **** -- ends -- */

/* routinint es in this file:

    void        codePanelEditorButtonCb (void)
    void        codePanelExitCb  (Fl_Widget *widget,void *)
    static void editorDone       (int signum)
    static int  execCommand      (char *command,char *errbuf,int buflen)
    char        *getFileContent  (char *path)
    static int  parseArgs        (char *buf,char **args,int args_len,int *n)
    void        startEditor      (char *command,char *tmpfile)
    void        tempFilename     (char *filename,int len)

    staic const char *get_code_input   (void)
    void  set_code_input               (const char *txt)
*/

/*
** deactivate code input text area, Editor..., ok and cancel button
** after the editor is started
*/
static void deActivateButtons(void)
{
	code_input->deactivate();
    code_panel_editor->deactivate();
    code_panel_ok->deactivate();
    code_panel_cancel->deactivate();
}

/*
** activate code input text area, Editor..., ok and cancel button
** after the editor is ended.
*/
static void activateButtons(void)
{
	code_input->activate();
    code_panel_editor->activate();
    code_panel_ok->activate();
    code_panel_cancel->activate();
}

/*
**  get_code_input()
**      Gets the current contents of code_input.
**
**  Parameters:
**      none
**
**  Return Values:
**      The contents of the code_input. User must call free() when done.
**
**  Limitations and Comments:
**      Fluid 1.3 has code_input as a Fl_Text_Editor.
**      This also makes porting this to future versions of
**      fluid easier as it requires this to be only modified
**      if the code_panel widget changes.
**
**  Development History:
**      who                      when            why
**      alvinbeach@hotmail.com   Jan-30-2005     Initially created
*/
static const char *get_code_input()
{
	return code_input->value();
}

/*
**  set_code_input()
**      Sets the contents of code_input.
**
**  Parameters:
**      The text to write into the code_input
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      Fluid 1.3 has code_input as a Fl_Text_Editor.
**      This also makes porting this to future versions of
**      fluid easier as it requires this to be only modified
**      if the code_panel widget changes.
**
**  Development History:
**      who                      when            why
**      alvinbeach@hotmail.com   Jan-30-2005     Initially created
*/
static void set_code_input(const char *txt)
{
	fprintf(stderr, "Setting text: %s\n", txt);
	code_input->value(txt);
}


/*
**  writeCodePanelStatus()
**
**  Parameters:
**      char *lab - the NULL terminated message
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      A copy of the message is made before setting to the label.
**      A static var is used to copy the message, so it frees it if it
**      already points to a malloc'd space.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    May-20-2001    first cut
*/
/*void writeCodePanelStatus(char *lab)
{
    static char
        *s=NULL;

    if (s != NULL)
        (void) free((char *) s);

    s=strdup(lab);
    if (s != NULL)
    {
        code_panel_status->label(s);
        code_panel_status->redraw();
    }
}
*/

/*
**  doBeep()
**      beep
**
**  Parameters:
**      none
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      for Unix and Win32
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-17-2001     first cut
*/
static void doBeep(void)
{
#ifdef WIN32
            MessageBeep(MB_ICONERROR);
#else /* UNIX */
            if (!fl_display)
                fl_open_display();
            if (fl_display)
                XBell(fl_display, 100);
            else
            {
                printf("\007"); /* doesn't work in windoze */
                (void) fflush(stdout);
            }
#endif /* UNIX */
    return;
}

#ifdef WIN32
/*
**  editorThread()
**      The routine called from the thread to start the editor. It creates
**      the editor process, if the process is created successfully:
**          - it wil deactivate the code input text area, the button in
**            the code panel.
**          - will for the editor process to finish and after that will
**            activate the text area and the buttons.
**
**  Parameters:
**      void *cmd - the editor command (e.g. cmd /c vim)
**
**  Return Values:
**      NULL
**
**  Limitations and Comments:
**      For Win32 only. Simulation of Unix fork()/exec()
**      the tmpfile will be removed.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    May-23-2001    first cut
*/
unsigned __stdcall editorThread(void *cmd)
{
    PROCESS_INFORMATION
        process_info;

    STARTUPINFO
        startup_info;

    BOOL
        ec,
        rc;
    
    DWORD
        dw_exit_code,
        dw;

    HANDLE
        h_process,
        h_thread;

    char
        *code;

    char
        command[BUFSIZ];

    memset(&startup_info,0,sizeof(startup_info));
    startup_info.cb=sizeof(startup_info);

#ifdef DEBUG
    (void) fprintf(stderr," cmd=%s\n",cmd);
#endif /* DEBUG */


    rc=CreateProcess(NULL,
            (char *) cmd,
            NULL,
            NULL,
            FALSE,
            NORMAL_PRIORITY_CLASS,
            NULL,
            NULL,
            &startup_info,
            &process_info);

    if (rc)
    {
        h_process=process_info.hProcess;
        h_thread=process_info.hThread;

        deActivateButtons();

        dw=WaitForSingleObject(h_process,INFINITE);
        if (dw != 0xffffffff) /* Success */
        {
            ec=GetExitCodeProcess(h_process,&dw_exit_code);
#ifdef DEBUG
            (void) fprintf(stderr," ec=%d exit_code=%d\n",
                           ec,dw_exit_code);
#endif /* DEBUG */
        }
        CloseHandle(h_thread);
        CloseHandle(h_process);

        code=getFileContent(s_tmpfile);
        if (code != NULL)
        {
			set_code_input(code);
            /*
            ** according to the doc the above calls copies the string
            ** to its internal buffer, so we should be able to free
            ** our malloc'd space.
            ** muquit@muquti.com May-23-2001 
            */
            (void) free((char *) code); /* XXX */
        }
        unlink(s_tmpfile);

        writeCodePanelStatus(" ");
        code_panel->cursor(FL_CURSOR_DEFAULT);
        s_editor_running=0;
        activateButtons();
    }
    else
    {
        (void) fprintf(stderr,"Could not run: %s\n",cmd);
    }

    _endthreadex(0);
    return (NULL);
}
#endif /* WIN32 */

/*
**  getFileContent()
**      opens a file, read the content and returns a pointer to it.
**
**  Parameters:
**      char *path      - the full path of the file 
**
**  Return Values:
**      char *content   - on success
**      NULL            - on failure
**
**  Limitations and Comments:
**      The content is pointer to a malloc'd space, the caller responsible
**      to free it. The content is NULL terminated. The file is closed.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-09-2001     first cut
**      muquit@muquit.com   Jun-11-2001     fixed an ABR
*/
char *getFileContent(char *path)
{
    char
        *whole_file=NULL;

    int
        fd,
        rc=0;

    struct stat
        sbuf;

    size_t
        count,
        length;

#ifdef WIN32
    HANDLE
        fh;
    DWORD
        nIn;
#endif /* WIN32 */

    rc=stat(path,&sbuf);
    if (rc != 0)
    {
        (void) fprintf(stderr,"could not stat file: %s\n",path);
        return (NULL);
    }
    length=sbuf.st_size;
    if (length == 0)
        return (NULL);

#ifdef WIN32
    fh=CreateFile(path,
            GENERIC_READ,
            0,
            NULL,
            OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL,
            NULL);
    if (fh == NULL)
    {
        (void) fprintf(stderr,"Could not open file %s for reading\n",path);
        return (NULL);
    }
#else /* UNIX */
    fd=open(path,O_RDONLY,0777);
    if (fd < 0)
    {
        (void) fprintf(stderr,"Could not open file: %s for reading\n",path);
        return (NULL);
    }
#endif /* ! WIN32 (UNIX) */

    whole_file=(char *) malloc(length+1);
    if (whole_file == (char *) NULL)
    {
        (void) fprintf(stderr,"memory allocation failed for whole_file\n");
#ifdef WIN32
        CloseHandle(fh);
#else /* UNIX */
        close(fd);
#endif /* ! WIN32 (UNIX) */
        return (NULL);
    }
    memset(whole_file,0,length+1);

#ifdef WIN32
    ReadFile(fh,whole_file,length,&nIn,NULL);    
    if (length != nIn)
    {
        (void) fprintf(stderr,"Error reading file: %s\n",path);
        CloseHandle(fh);
        return (NULL);
    }
#else /* UNIX */
    count=read(fd,whole_file,(length*sizeof(char))+1);
    if (count != length)
    {
        (void) fprintf(stderr,"Coult not read file: %s\n",path);
        close(fd);
        (void) free((char *) whole_file);
        return (NULL);
    }

    close(fd);

#endif /* ! WIN32 (UNIX) */

    if (whole_file[length] != '\0')
    {
#ifdef DEBUG
        (void) fprintf(stderr,"getFileContent() - NULL terminated at %d \n",length);
#endif /* DEBUG */
        whole_file[length]='\0';
    }
    return (whole_file);
}

#ifndef WIN32
/*
**  editorDone()
**      called when the editor is exited or killed
**
**  Parameters:
**      int signum  - the signal 
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      called as the signal handler of SIGCHLD. is called when the child
**      dies. This routine is for Unix only.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-17-2001     first cut
**      muquit@muquit.com   Jun-11-2001     fixed a memory leak
**      alvinbeach@hotmail.com Jan-30-2005  Updated to call set_code_input()
*/
static void editorDone(int signum)
{
    int
        pid;

    char
        *code;

    pid=waitpid(-1,(int *) 0,WNOHANG);
    if (pid == -1)
        return;

#ifdef DEBUG
    (void) fprintf(stderr,"editorDone() - process %d ended\n",pid);
#endif /* DEBUG */

    activateButtons();

    code=getFileContent(s_tmpfile);
    if (code != NULL)
    {
		set_code_input(code);
        (void) free((char *) code);
    }
    unlink(s_tmpfile);

    //writeCodePanelStatus(" ");
    code_panel->cursor(FL_CURSOR_DEFAULT);
    s_editor_running=0;
}
#endif /* ! WIN32 (UNIX) */


/*
**  parseArgs()
**      split commands to NULL terminated individual args
**
**  Parameters:
**      char    *buf     - the command
**      char    **args   - NULL terminated args, returns
**      int     args_len - number of pointer args can hold
**      int     *n       - number of args in the command, returns
**
**  Return Values:
**      0   on success
**      -1 on failure
**
**  Limitations and Comments:
**      command can have 128 args max. This routine is used in Unix only.
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-17-2001     first cut
*/
static int parseArgs(char *buf,char **args,int args_len,int *n)
{
    *n=0;
    while (*buf != '\0')
    {
        while ((*buf == ' ')  ||
               (*buf == '\t') ||
               (*buf == '\n'))
        {
            *buf++='\0';
        }
        if (++(*n) > args_len)
        {
            (void) fprintf(stderr,"args buffer overflowed: %d > % d\n",
                           *n,args_len);
            return(-1);
        }
        *args++=buf;

        /* skip over the argument */
        while ((*buf != '\0') &&
               (*buf != ' ')  &&
               (*buf != '\t') &&
               (*buf != '\n'))
            buf++;
    }
    *args='\0';
    return (0);
}

/*
**  execCommand()
**      executes a external command
**
**  Parameters:
**      char *command   - the command to execute
**      char *errbuf    - returns error message if any
**      int  buflen     - length of errbuf
**
**  Return Values:
**      0   on success
**      -1  on failure
**
**  Limitations and Comments:
**      For Unix and Win32. For Win32, it starts a new thread and start
**      the editor from the thread handler routine and the handler routine
**      waits till the editor is done (In Unix, the signal handler does
**      that).
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    May-15-2001    first cut
*/
static int execCommand(char *command,char *errbuf,int buflen)
{
    int
        pid,
        rc,
        n=0;

    char
        *args[128],
        cmd_buf[BUFSIZ];

#ifdef WIN32
    HANDLE
        thread_handle;

    DWORD
        ret_val;

    unsigned int
        thread_id;
#endif /* WIN32 */

    (void) strcpy(cmd_buf,command);
#ifndef WIN32 /* UNIX */
    rc=parseArgs(cmd_buf,args,127,&n);
    if (rc != 0)
    {
        (void) fprintf(stderr,"Could not parse:\"%s\"\n",command);
        (void) sprintf(errbuf,"Could not parse:\"%s\"",command);
        return (-1);
    }

#ifdef DEBUG
    if (n)
    {
        for (int i=0; i < n; i++)
        {
            (void) fprintf(stderr,"execCommand() - args[%d]=%s\n",i,args[i]);
        }
    }
#endif /* DEBUG */
#endif /* UNIX */

#ifdef WIN32
    /* start a thread */
    thread_handle=(HANDLE) _beginthreadex(NULL,
            0,
            editorThread,
            (void *) cmd_buf,
            0,
            &thread_id);
#else /* UNIX */

    pid=fork();
    switch(pid)
    {
        case -1:
        {
            (void) strcpy(errbuf,"fork() failed out of process\r\n");
            return(-1);
            break;

        }
        case 0: /* child */
        {
            execvp(args[0],args);
            (void) sprintf(errbuf,"failed to exec: %s",command);
            perror("execvp");
            (void) fprintf(stderr," could not exec: %s\n",command);
            _exit(255);
            break;
        }

        default:
        {
            break;
        }
    }

#endif /* ! WIN32 (UNIX) */

    return(0);
}



/*
**  tempFilename()
**      make a file with unique name
**
**  Parameters:
**      char *filename - filled with the name (returns)
**      int len        - lenth of text filename can hold
**
**  Return Values:
**      0 on success
**      -1 on failure
**
**  Limitations and Comments:
**      Set the TMPDIR to your own space in Unix for security reason.
**      I add .cxx at the end of the filename because I use vim and I
**      want the syntax highlighting gets turned on automatically.
**
**      ARB: tempnam is deprecated but mkstemp doesn't allow us to append .cxx
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-13-2001     first cut
*/
int tempFilename(char *filename,int len)
{
#define TEMP_TEMPLATE       "fltk"
    char
        *name;

    name=(char *) tempnam((char *) NULL,TEMP_TEMPLATE);
    if (name == NULL)
    {
        return (-1);
    }
    else
    {
        (void) strncpy(filename,name,len-10);
        filename[len-10]='\0';
        (void) strcat(filename,".cxx");
        (void) free((char *) name);
    }

    return (0);
}


/*
**  startEditor()
**      start editor 
**
**  Parameters:
**      char *command - the command, it might look like:
**                      xterm -e vim fltk353xr.cxx
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**
**
**
**  Development History:
**      who                     when          why
**      muquit@muquit.com       May-13-2001   first cut
**      alvinbeach@hotmail.com  Jan-30-2005   Commented writing to status bar
**                                            since original fluid does have one
*/
void startEditor(char *command,char *tmpfile)
{
    char
        errbuf[BUFSIZ],
        buf[BUFSIZ];

    int
        rc;
            
    (void) strcat(command," ");
    (void) strcat(command,s_tmpfile);
    rc=execCommand(command,errbuf,sizeof(errbuf)-1);
    if (rc != 0)
    {
        fl_alert(errbuf);
        return;
    }
    
    deActivateButtons();

    s_editor_running=1;

    (void) sprintf(buf,"Editing in file: %s",tmpfile);

#ifdef DEBUG
    fl_message(buf);
#endif
    //writeCodePanelStatus(buf);
    //code_panel->cursor(FL_CURSOR_WAIT);
}



/*
**  codePanelHelpButtonCb()
**      callback routine for the Help button in code panel
**
**  Parameters:
**      none
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      the help window must be created first.
**
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com    May-22-2001    first cut
*/

void codePanelHelpButtonCb(void)
{
    code_editor_help_win->show();
}


/*
**  codePanelEditorButtonCb()
**    start editor with the tmp file
**
**  Parameters:
**      none
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      uses the statics  in module: s_tmpfile
**      uses globals: code_input
**      if the environment variable FLTK_EDITOR does not exist, return
**
**  Development History:
**      who                     when          why
**      muquit@muquit.com       May-13-2001   first cut
**      muquit@muquit.com       May-21-2001   Win32 port
**      alvinbeach@hotmail.com  Jan-30-2005   Added extra check if FLUID_EDITOR
**                                            had zero length.
*/
void codePanelEditorButtonCb(void)
{
    char
        buf[BUFSIZ],
        *s;

    const char
        *code;

    int
        rc;

#ifdef WIN32
    HANDLE
        hwfile;
    DWORD
        nOut;
#else /* UNIX */

    /* if you hvae SIGACTION */
#define HAVE_SIGACTION 1

#ifdef HAVE_SIGACTION
    struct sigaction
        act;
#endif /* HAVE_SIGACTION */

#endif /* WIN32 */

    s=getenv("FLUID_EDITOR");
    if ((s == NULL) || (strlen(s) == 0))
    {
        doBeep();

        fl_alert("FLUID_EDITOR environment variable is not set!");
		codePanelHelpButtonCb();
        return;
    }
    (void) strncpy(buf,s,sizeof(buf)-1);
    buf[sizeof(buf)-1]='\0';


    /* create a tmp filename */
    memset(s_tmpfile,0,sizeof(s_tmpfile));
    rc=tempFilename(s_tmpfile,sizeof(s_tmpfile)-1);
    if (rc != 0)
    {
        fl_alert("Could not create tmp file");
        return;
    }

    /* if one already exits, get rid of it */
    unlink(s_tmpfile);

    /* get the code from code_input text window */
	code = get_code_input();

    /* put the code in the tmp file */
#ifdef WIN32
    if (code != NULL)
    {
        hwfile=CreateFile(s_tmpfile,
                GENERIC_WRITE,
                0,
                NULL,
                CREATE_ALWAYS,
                FILE_ATTRIBUTE_NORMAL,
                NULL);

        if (hwfile != NULL)
        {
            WriteFile(hwfile,code,(DWORD) strlen(code),&nOut,NULL);
            CloseHandle(hwfile);
        }

    }
#else /* UNIX */
    if (code != NULL)
    {
        FILE
            *fp;

        fp=fopen(s_tmpfile,"wb");
        if (fp != NULL)
        {
            (void) fprintf(fp,"%s",code);
#ifdef DEBUG
			fprintf(stderr, "Writing to tmp file: %s\n", code);
#endif
            (void) fclose(fp);
        }
    }
	else
		fprintf(stderr, "Code input was empty\n");

	// ALVIN: In fltk 1.1.6 value() returns a copy. In ftk 2.0 it doesn't!
	//free((char*)code);

    /*
    ** we'll fork a process and start editor, setup the signal handler
    ** so that the parent knows when the child ends
    */
    sigemptyset(&act.sa_mask);
    act.sa_flags=0;
    act.sa_handler=editorDone;
    sigaction(SIGCHLD,&act,NULL);

    /* signal(SIGCHLD,editorDone); */ /* old way */
#endif /* !WIN32 (UNIX) */


    /* start the editor */
    startEditor(buf,s_tmpfile);
}

/*
**  codePanelExitCb()
**    exit callback of the code panel window
**
**  Parameters:
**      Fl_Wiget *  - not used
**      void *      - not used
**
**  Return Values:
**      none
**
**  Limitations and Comments:
**      args are not used at this time      
**
**  Development History:
**      who                  when           why
**      muquit@muquit.com   May-13-2001     first cut
*/
void codePanelExitCb(Fl_Widget *widget,void *)
{
    if (! s_editor_running)
    {
        code_panel->hide();
    }
    else
        doBeep();
}

